# -*- coding: utf-8 -*-
"""
This module is responsible for installing the backend for the C++ plugin.
"""
__all__ = ('post_install,')

import contextlib
import pkg_resources

from loguru import logger
import docker

IMAGE_NAME: str = 'christimperley/kaskara:cpp'


def post_install() -> None:
    """Installs the C++ plugin backend."""
    logger.info('installing C++ plugin backend')
    backend_directory = pkg_resources.resource_filename(__name__, 'backend')
    logger.debug(f'backend located at: {backend_directory}')
    with contextlib.closing(docker.from_env()) as docker_client:
        logger.debug('prepared Docker client for C++ backend installation')
        image, _ = docker_client.images.build(path=backend_directory,
                                              tag=IMAGE_NAME,
                                              pull=True)
        logger.info(f"built Docker image for C++ plugin: {image.tags}")
    logger.info('installed C++ plugin backend')



# -*- coding: utf-8 -*-
__all__ = ('SpoonAnalyser',)

from typing import Any, Iterator, List, Mapping, Sequence
import contextlib
import json
import os
import shlex
import subprocess

from dockerblade import DockerDaemon as DockerBladeDockerDaemon
from loguru import logger
import attr

from .analysis import SpoonFunction, SpoonStatement
from .post_install import IMAGE_NAME as SPOON_IMAGE_NAME
from ..analyser import Analyser
from ..analysis import Analysis
from ..container import ProjectContainer
from ..core import FileLocationRange
from ..functions import ProgramFunctions
from ..loops import ProgramLoops
from ..project import Project
from ..statements import ProgramStatements


@attr.s
class SpoonAnalyser(Analyser):
    _dockerblade: DockerBladeDockerDaemon = attr.ib(repr=False)

    @contextlib.contextmanager
    def _container(self, project: Project) -> Iterator[ProjectContainer]:
        """Provisions an ephemeral container for a given project."""
        launch = self._dockerblade.client.containers.run
        with contextlib.ExitStack() as stack:
            # create a temporary volume from the project image
            volume_name = 'kaskaraspoon'
            cmd_create_volume = (f'docker run --rm -v {volume_name}:'
                                 f'{shlex.quote(project.directory)} '
                                 f'{project.image} /bin/true')
            cmd_kill_volume = f'docker volume rm {volume_name}'
            logger.debug(f'created temporary volume [{volume_name}] '
                         f'from project image [{project.image}] '
                         f'via command: {cmd_create_volume}')
            subprocess.check_output(cmd_create_volume, shell=True)
            stack.callback(subprocess.call, cmd_kill_volume,
                           shell=True,
                           stderr=subprocess.DEVNULL,
                           stdout=subprocess.DEVNULL,
                           stdin=subprocess.DEVNULL)

            docker_analyser = launch(SPOON_IMAGE_NAME, '/bin/sh',
                                     stdin_open=True,
                                     volumes={volume_name: {
                                         'bind': '/workspace',
                                         'mode': 'ro'}},
                                     detach=True)
            stack.callback(docker_analyser.remove, force=True)

            dockerblade = self._dockerblade.attach(docker_analyser.id)
            yield ProjectContainer(project=project, dockerblade=dockerblade)

    def analyse(self, project: Project) -> Analysis:
        logger.debug(f"analysing Spoon project: {project}")
        with self._container(project) as container:
            return self._analyse_container(container)

    def _analyse_container(self, container: ProjectContainer) -> Analysis:
        dir_source = '/workspace'
        dir_output = '/output'
        container.shell.check_output(f'kaskara {dir_source} -o {dir_output}')

        # load statements
        filename_statements = os.path.join(dir_output, 'statements.json')
        statements_dict = json.loads(container.files.read(filename_statements))
        statements = self._load_statements_from_dict(statements_dict)

        # load functions
        filename_functions = os.path.join(dir_output, 'functions.json')
        functions_dict = json.loads(container.files.read(filename_functions))
        functions = self._load_functions_from_dict(functions_dict)

        # load loops
        filename_loops = os.path.join(dir_output, 'loops.json')
        loops_dict = json.loads(container.files.read(filename_loops))
        loops = self._load_loops_from_dict(loops_dict)

        # find insertion points
        insertions = statements.insertions()

        return Analysis(project=container.project,
                        loops=loops,
                        functions=functions,
                        statements=statements,
                        insertions=insertions)

    def _load_statements_from_dict(self,
                                   dict_: Sequence[Mapping[str, Any]]
                                   ) -> ProgramStatements:
        """Loads the statement database from a given dictionary."""
        logger.debug('parsing statements database')
        statements = \
            ProgramStatements([SpoonStatement.from_dict(d) for d in dict_])
        logger.debug(f'parsed {len(statements)} statements')
        return statements

    def _load_functions_from_dict(self,
                                  dict_: Sequence[Mapping[str, Any]]
                                  ) -> ProgramFunctions:
        """Loads the function database from a given dictionary."""
        logger.debug('parsing function database')
        functions = \
            ProgramFunctions([SpoonFunction.from_dict(d) for d in dict_])
        logger.debug(f'parsed {len(functions)} functions')
        return functions

    def _load_loops_from_dict(self,
                              dict_: Sequence[Mapping[str, Any]]
                              ) -> ProgramLoops:
        """Loads the loops database from a given dictionary."""
        logger.debug('parsing loop database')
        loop_bodies: List[FileLocationRange] = []
        for loop_info in dict_:
            loc = FileLocationRange.from_string(loop_info['body'])
            loop_bodies.append(loc)
        loops = ProgramLoops.from_body_location_ranges(loop_bodies)
        logger.debug(f'parsed loops')
        return loops



# -*- coding: utf-8 -*-
"""
This module is used to install Kaskara's various language backends following
the installation of the Kaskara Python package via setuptools/pip.
"""
__all__ = ('post_install',)

from loguru import logger

from .clang import post_install as post_install_clang


def post_install() -> None:
    logger.info('performing post-installation of Kaskara')
    post_install_clang()
    logger.info('completed post-installation of Kaskara')



# -*- coding: utf-8 -*-
__all__ = ('ClangFunction', 'ClangStatement')

from typing import Any, FrozenSet, Mapping

import attr
from loguru import logger

from ..core import FileLocationRange
from ..functions import Function
from ..project import Project
from ..statements import Statement
from ..util import abs_to_rel_flocrange


@attr.s(frozen=True, slots=True, auto_attribs=True)
class ClangFunction(Function):
    name: str
    location: FileLocationRange
    body_location: FileLocationRange
    return_type: str
    is_global: bool
    is_pure: bool

    @staticmethod
    def from_dict(project: Project, d: Mapping[str, Any]) -> 'ClangFunction':
        name = d['name']
        location = FileLocationRange.from_string(d['location'])
        location = abs_to_rel_flocrange(project.directory, location)
        body = FileLocationRange.from_string(d['body'])
        body = abs_to_rel_flocrange(project.directory, body)
        return_type = d['return-type']
        is_global = d['global']
        is_pure = d['pure']
        return ClangFunction(name=name,
                             location=location,
                             body_location=body,
                             return_type=return_type,
                             is_global=is_global,
                             is_pure=is_pure)


@attr.s(frozen=True, slots=True, auto_attribs=True)
class ClangStatement(Statement):
    content: str
    canonical: str
    kind: str
    location: FileLocationRange
    reads: FrozenSet[str]
    writes: FrozenSet[str]
    visible: FrozenSet[str]
    declares: FrozenSet[str]
    live_before: FrozenSet[str]
    live_after: FrozenSet[str]
    requires_syntax: FrozenSet[str]

    @staticmethod
    def from_dict(project: Project, d: Mapping[str, Any]) -> 'ClangStatement':
        location = FileLocationRange.from_string(d['location'])
        location = abs_to_rel_flocrange(project.directory, location)
        statement = ClangStatement(d['content'],
                                   d['canonical'],
                                   d['kind'],
                                   location,
                                   frozenset(d.get('reads', [])),
                                   frozenset(d.get('writes', [])),
                                   frozenset(d.get('visible', [])),
                                   frozenset(d.get('decls', [])),
                                   frozenset(d.get('live_before', [])),
                                   frozenset(d.get('live_after', [])),
                                   frozenset(d.get('requires_syntax', [])))
        logger.debug(f"loaded statement: {statement}")
        return statement
